unit anADODatPrv;

(*******************************************************************************
v1.1

This component is the result of a project where I needed access to MS SQL
Server with ADO, and is provided as is. It works in my project, but I don't make
any guarantees that it will work for you. You may use this component on your own
risk. I don't take any responsibility for any problems arising from the use
of it (including but not limited to data loss).

Se readme.txt for help with this component and license.txt for conditions of
use.

Anders Nolberger
Sekonden IT AB
Teknikringen 6
583 30 Linkping
Sweden

tel: +46 13 214440
fax: +46 13 214441
Email: Anders.Nolberger@sekonden.se
WEB: HTTP://www.sekonden.se

*******************************************************************************)

interface

uses
{$IFDEF WIN32}
  Windows,
{$ELSE}
  WinProcs, WinTypes,
{$ENDIF}

  Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  Grids, DB, DBGrids, StdCtrls, DBCtrls, ExtCtrls, Buttons,
  ppTypes, ppClass, ppDB, ppUtils, daDataVw,
  ADODB_TLB, anADOPipeline;

type

  TanADOPreviewDataDialog = class(TdaCustomPreviewDataDialog)
    Panel1: TPanel;
    btnOK: TButton;
    StringGrid1: TStringGrid;
    lblNumOfRecords: TLabel;

  private
    FConnectionString: String;
    FSQLQuery: String;
    function GetRealConnectionstring(Databasename: String): String;

  protected
    {override from ancestor}
    procedure SetDataView(aDataView: TdaCustomDataView); override;
    procedure LanguageChanged; override;
    procedure FillGrid;

  end;

var
  anADOPreviewDataDialog: TanADOPreviewDataDialog;

implementation

{$R *.DFM}


{------------------------------------------------------------------------------}
{ TanADOPreviewDataDialog.SetDataView }

procedure TanADOPreviewDataDialog.SetDataView(aDataView: TdaCustomDataView);
begin

  inherited SetDataView(aDataView);


  if (aDataView <> nil) then
    begin
      aDataView.Active := True;
      if (aDataView.DataPipelineCount > 0) then

        {check whether DataPipeline has a ConnectionString property}
        if ppCheckPropInfo(aDataView.DataPipelines[0],  'ConnectionString') then
          FConnectionString := ppGetStringPropValue(aDataView.DataPipelines[0], 'ConnectionString');

        {check whether DataPipeline has a SQLQuery property}
        if ppCheckPropInfo(aDataView.DataPipelines[0],  'SQLQuery') then
          FSQLQuery := ppGetStringPropValue(aDataView.DataPipelines[0], 'SQLQuery');
    end;

  LanguageChanged;

  FillGrid;

  if aDataView = nil then Exit;

  Caption := ppLoadStr(808) + ' - ' + DataView.Description; {Preview Data}


end; {procedure, SetDataView}

{------------------------------------------------------------------------------}
{ TanADOPreviewDataDialog.LanguageChanged; }

procedure TanADOPreviewDataDialog.LanguageChanged;
begin

  if DataView <> nil then
    Caption := ppLoadStr(808) + ' - ' + DataView.Description {Preview Data}
  else
    Caption := ppLoadStr(808); {Preview Data}


  btnOK.Caption := ppLoadStr(ppMsgOK); {OK}

end; {procedure, LanguageChanged;}

procedure TanADOPreviewDataDialog.FillGrid;
var
  cn: Connection;
  rs: Recordset;
  i, Col, Row: Integer;
  MaxRow: Integer;
begin
  try
    cn := CoConnection.Create;
    if Assigned(cn) then
    begin
      cn.CursorLocation := adUseClient;
      cn.ConnectionTimeout := 10;
      cn.Open(GetRealConnectionstring(FConnectionString), '', '', 0);
      if cn.State = adStateOpen then
      begin
        rs := CoRecordset.Create;
        if Assigned(rs) then
        begin
          rs.Open(FSQLQuery, cn, adOpenStatic, adLockReadOnly, adCmdText);
        end;
      end;
    end;

    if (not rs.BOF) and (not rs.EOF) then
    begin
      StringGrid1.ColCount := rs.Fields.Count;
      for i := 0 to rs.Fields.Count - 1 do
      begin
        StringGrid1.Cells[i, 0] := rs.Fields.Item[i].Name;
        StringGrid1.ColWidths[i] := Length(rs.Fields.Item[i].Name) * 10;
      end;
    end;

    if rs.RecordCount > 200 then
    begin
      MaxRow := 200;
      lblNumOfRecords.Caption := 'Showing 200 records of total ' + IntToStr(rs.Recordcount) + '.';
    end
    else
    begin
      MaxRow := rs.RecordCount;
      lblNumOfRecords.Caption := IntToStr(rs.Recordcount) + ' records.';
    end;
    if rs.RecordCount = 0 then
    begin
      StringGrid1.RowCount := 2;
      for col := 0 to StringGrid1.ColCount - 1 do
      begin
        StringGrid1.Cells[col, 1] := '';
      end;
    end
    else
    begin
      StringGrid1.RowCount := MaxRow + 1;
    end;
    for row := 1 to MaxRow do
    begin
      for col := 0 to rs.Fields.Count - 1 do
      begin
        try
          case rs.Fields.Item[Col].Type_ of
            adBSTR, adChar, adVarChar, adLongVarChar, adWChar, adVarWChar, adLongVarWChar:
            begin
              StringGrid1.Cells[Col, Row] := rs.Fields.Item[Col].Value;
            end;
            adDBTime, adDBTimeStamp, adDate, adDBDate:
            begin
              StringGrid1.Cells[Col, Row] := rs.Fields.Item[Col].Value;
            end;
            adVarNumeric, adDouble, adDecimal, adNumeric, adCurrency, adSingle,
            adBigInt, adUnsignedBigInt, adTinyInt, adSmallInt, adInteger, adUnsignedTinyInt,
            adUnsignedSmallInt, adUnsignedInt:
            begin
              StringGrid1.Cells[Col, Row] := rs.Fields.Item[Col].Value;
            end;
            adGUID:
            begin
              StringGrid1.Cells[Col, Row] := rs.Fields.Item[Col].Value;
            end;
            adBoolean:
            begin
              if rs.Fields.Item[Col].Value then
              begin
                 StringGrid1.Cells[Col, Row] := 'True';
              end
              else
              begin
                 StringGrid1.Cells[Col, Row] := 'False';
              end;
            end;
          else
          end;
        except
          StringGrid1.Cells[Col, Row] := '';
        end;
      end;
      rs.MoveNext;
    end;
  except
  end;

  if Assigned(rs) then
  begin
    if rs.State = adStateOpen then
    begin
      rs.Close;
    end;
    rs := nil;
  end;
  // ADO
  if Assigned(cn) then
  begin
    if cn.State = adStateOpen then
    begin
      cn.Close;
    end;
    cn := nil;
  end;
end;

function TanADOPreviewDataDialog.GetRealConnectionstring(Databasename: String): String;
var
  i, j: Integer;
  Form: TForm;
  ConStr: String;
begin
  ConStr := Databasename;
  for i := 0 to Application.ComponentCount - 1 do
  begin
    if Application.Components[i] is TForm then
    begin
      Form := TForm(Application.Components[i]);
      for j := 0 to Form.ComponentCount - 1 do
      begin
        if Form.Components[j].ClassName = 'TanADOPipeline' then
        begin
          if TanADOPipeline(Form.Components[j]).UserName = Databasename then
          begin
            ConStr := TanADOPipeline(Form.Components[j]).Connectionstring;
          end;
        end;
      end;
    end;
  end;
  Result := ConStr;
end;

initialization

  RegisterClass(TanADOPreviewDataDialog);

finalization

  UnRegisterClass(TanADOPreviewDataDialog);

end.
