unit anADOPipeline;

(*******************************************************************************
v 1.1

This component is the result of a project where I needed access to MS SQL
Server with ADO, and is provided as is. It works in my project, but I don't make
any guarantees that it will work for you. You may use this component on your own
risk. I don't take any responsibility for any problems arising from the use
of it (including but not limited to data loss).

Se readme.txt for help with this component and license.txt for conditions of
use.

Anders Nolberger
Sekonden IT AB
Teknikringen 6
583 30 Linkping
Sweden

tel: +46 13 214440
fax: +46 13 214441
Email: Anders.Nolberger@sekonden.se
WEB: HTTP://www.sekonden.se

*******************************************************************************)

interface

uses
  ADODB_TLB, Classes, SysUtils, Graphics, Dialogs, Windows, ComObj, Forms,
  ppTypes, ppDB;

type
  TanADOPipeline = class(TppCustomDataPipeline)
  private
    // ADO
    FConnection: Connection;
    // RDS
    FDataSpace: Variant;
    FDataFactory: Variant;
    //Common
    FRecordset: Variant;

    FConnectionString: String;
    FConStringIsShortcut: Boolean;
    FSQLQuery: String;
    FServer: String;

    FTemplate: Boolean;
    
    procedure SetConnectionString(Value: String);
    procedure SetConStringIsShortcut(Value: Boolean);
    procedure SetSQLQuery(Value: String);
    procedure SetServer(Value: STring);
    procedure SetActive(Value: Boolean);
  protected
    procedure OpenDataSet; override;
    procedure CloseDataSet; override;
    function  GetActive: Boolean; override;
    function  CheckBOF: Boolean; override;
    function  CheckEOF: Boolean; override;
    procedure GotoFirstRecord; override;
    procedure GotoLastRecord; override;
    procedure TraverseBy(aIncrement: Integer); override;
  public
    constructor Create(aOwner: TComponent); override;
    destructor Destroy; override;

    procedure RestoreFromCache(aCachePageNo: Longint); override;
    procedure SaveToCache(aCachePageNo: Longint); override;

    function  GetBookmark: Longint; override;
    procedure GotoBookmark(aBookmark: Longint); override;
    procedure FreeBookmark(aBookmark: Longint); override;
    function  GetDataSetName: String; override;
    function  GetFieldNames(aFieldNameList: TStrings): Boolean; override;
    function  GetFieldDataType(aFieldName: String): TppDataType; override;
    function  GetFieldSize(aFieldName: String): Integer; override;
    function  GetFieldDisplayWidth(aFieldName: String): Integer; override;
    function  GetFieldIsCalculated(aFieldName: String): Boolean; override;
    function  GetFieldAsDouble(aFieldName: String): Double; override;
    function  GetFieldAsPicture(aFieldName: String): TPicture; override;
    function  GetFieldAsString(aFieldName: String): String; override;
    function  GetFieldValue(aFieldName: String): Variant; override;

    procedure Delete; override;
    procedure Edit; override;
    procedure Insert; override;
    function  Locate(const aFieldName: string; aKeyValue: Variant; aOptions: TppLocateOptions): Boolean; override;
    procedure Post; override;
    procedure SetFieldFromStream(aFieldName: String; aStream: TStream); override;
    procedure GetFieldAsStream(aFieldName: String; aStream: TStream); override;
    procedure SetFieldValue(aFieldName: String; aValue: Variant); override;


  published
    property ConnectionString: String read FConnectionString write SetConnectionString;
    property ConStringIsShortcut: Boolean read FConStringIsShortcut write SetConStringIsShortcut default False;
    property SQLQuery: String read FSQLQuery write SetSQLQuery;
    property Server: String read FServer write SetServer;
    property Active: Boolean read GetActive write SetActive;
    property Template: Boolean read FTemplate write FTemplate default False;

    {TppDataPipeline properties}
    property AutoCreateFields;
    property MasterDataPipeline;
    property MoveBy;
    property RangeEnd;
    property RangeEndCount;
    property RangeBegin;
    property SkipWhenNoRecords;
    property UserName;
    property Visible;

    {TppDataPipeline events}
    property OnClose;
    property OnDataChange;
    property OnFirst;
    property OnGotoBookmark;
    property OnLast;
    property OnMasterRecordPositionChange;
    property OnNext;
    property OnOpen;
    property OnPrior;
    property OnRecordPositionChange;
    property OnTraversal;
  end;

  TanChildADOPipeline = class(TanADOPipeline)
  public
    function HasParent: Boolean; override;
  end;

procedure Register;

implementation

procedure Register;
begin
  RegisterComponents('RBuilder', [TanADOPipeline]);

  RegisterNoIcon([TanChildADOPipeline]);
end;

{ TanChildADOPipeline }

function TanChildADOPipeline.HasParent: Boolean;
begin
  Result := True;
end;

{ TanADOPipeline }

constructor TanADOPipeline.Create(aOwner: TComponent);
begin
  inherited Create(aOwner);

  FConnection := nil;
  FConStringIsShortcut := False;
  FRecordset := Null;
end;

destructor TanADOPipeline.Destroy;
begin
  Destroying;
  Close;
  inherited Destroy;
end;

procedure TanADOPipeline.SetConnectionString(Value: String);
begin
  CloseDataset;
  FConnectionString := Value;
end;

procedure TanADOPipeline.SetConStringIsShortcut(Value: Boolean);
begin
  CloseDataset;
  FConStringIsShortcut := Value;
end;

procedure TanADOPipeline.SetSQLQuery(Value: String);
begin
  CloseDataset;
  FSQLQuery := Value;
end;

procedure TanADOPipeline.SetServer(Value: String);
begin
  CloseDataset;
  FServer := Value;
end;

procedure TanADOPipeline.OpenDataSet;
var
  i, j: Integer;
  Form: TForm;
  ConStr: String;
  Srv: String;
begin
  // Close old connection
  CloseDataSet;
  // Is this pipeline child to another pipeline?
  ConStr := FConnectionString;
  Srv := FServer;
  if FConStringIsShortcut then
  begin
    for i := 0 to Application.ComponentCount - 1 do
    begin
      if Application.Components[i] is TForm then
      begin
        Form := TForm(Application.Components[i]);
        for j := 0 to Form.ComponentCount - 1 do
        begin
          if Form.Components[j].ClassName = 'TanADOPipeline' then
          begin
            if TanADOPipeline(Form.Components[j]).UserName = FConnectionString then
            begin
              ConStr := TanADOPipeline(Form.Components[j]).Connectionstring;
              Srv := TanADOPipeline(Form.Components[j]).Server;
            end;
          end;
        end;
      end;
    end;
  end;
  if (FConnectionString <> '') and (FSQLQuery <> '') then
  begin
    if Srv = '' then
    begin
      // ADO
      try
        FConnection := CoConnection.Create;
        if Assigned(FConnection) then
        begin
          FConnection.CursorLocation := adUseClient;
          FConnection.ConnectionTimeout := 10;
          FConnection.Open(ConStr, '', '', 0);
          if FConnection.State = adStateOpen then
          begin
            FRecordset := CoRecordset.Create;
            if VarType(FRecordset) = varDispatch then
            begin
              FRecordset.Open(FSQLQuery, FConnection, adOpenStatic, adLockOptimistic, adCmdText);
            end;
          end;
        end;
      except
        on E:Exception do
        begin
          MessageBeep(MB_ICONEXCLAMATION);
          MessageDlg(E.Message, mtError, [mbOk], 0);
          CloseDataSet;
        end
        else
        begin
          CloseDataSet;
        end;
      end;
    end
    else
    begin
      // RDS
      try
        FDataSpace := CreateOleObject('RDS.DataSpace');
        if Vartype(FDataSpace) = varDispatch then
        begin
          FDataFactory := FDataSpace.CreateObject('RDSServer.DataFactory', Srv);
          if Vartype(FDataFactory) = varDispatch then
          begin
            FRecordset := FDataFactory.Query(ConStr, FSQLQuery);
          end;
        end;
      except
        on E:Exception do
        begin
          MessageBeep(MB_ICONEXCLAMATION);
          MessageDlg(E.Message, mtError, [mbOk], 0);
          CloseDataSet;
        end
        else
        begin
          CloseDataSet;
        end;
      end;
    end;
  end;
end;

procedure TanADOPipeline.CloseDataSet;
begin
  if VarType(FRecordset) = varDispatch then
  begin
    if FRecordset.State = adStateOpen then
    begin
      FRecordset.Close;
    end;
    FRecordset := Null;
  end;
  // ADO
  if Assigned(FConnection) then
  begin
    if FConnection.State = adStateOpen then
    begin
      FConnection.Close;
    end;
    FConnection := nil;
  end;
  // RDS
  FDataFactory := Null;
  FDataSpace := Null;
end;

function TanADOPipeline.GetActive: Boolean;
begin
  if (VarType(FRecordset) = varDispatch) and (FRecordset.State <> adStateClosed) then
  begin
    Result := True;
  end
  else
  begin
    Result := False;
  end;
end;

procedure TanADOPipeline.SetActive(Value: Boolean);
begin
  if Value then
    OpenDataset
  else
    CloseDataset;
end;

function TanADOPipeline.CheckBOF: Boolean;
begin
  if (VarType(FRecordset) = varDispatch) and (FRecordset.State <> adStateClosed) then
  begin
    Result := FRecordset.BOF;
  end
  else
  begin
    Result := True;
  end;
end;

function TanADOPipeline.CheckEOF: Boolean;
begin
  if (VarType(FRecordset) = varDispatch) and (FRecordset.State <> adStateClosed) then
  begin
    Result := FRecordset.EOF;
  end
  else
  begin
    Result := True;
  end;
end;

procedure TanADOPipeline.GotoFirstRecord;
begin
  try
    if (VarType(FRecordset) = varDispatch) and (FRecordset.State <> adStateClosed) then
      FRecordset.MoveFirst;
  except
  end;
end;

procedure TanADOPipeline.GotoLastRecord;
begin
  try
    if (VarType(FRecordset) = varDispatch) and (FRecordset.State <> adStateClosed) then
      FRecordset.MoveLast;
  except
  end;
end;


procedure TanADOPipeline.TraverseBy(aIncrement: Integer);
begin
  try
    if (VarType(FRecordset) = varDispatch) and (FRecordset.State <> adStateClosed) then
    begin
      if FRecordset.EOF or FRecordset.BOF then
        Exit;
      FRecordset.Move(aIncrement, adBookmarkCurrent);
    end;
  except
  end;
end;

function TanADOPipeline.GetBookmark: Longint;
begin
  if (VarType(FRecordset) = varDispatch) and (FRecordset.State <> adStateClosed) then
  begin
    try
      Result := FRecordset.BookMark;
    except
      Result := 0;
    end;
  end
  else
  begin
    Result := 0;
  end;
end;

procedure TanADOPipeline.GotoBookmark(aBookmark: LongInt);
var
  temp: Double;
begin
  temp :=aBookmark;
  if (VarType(FRecordset) = varDispatch) and (FRecordset.State <> adStateClosed) then
  begin
    try
      FRecordset.BookMark := temp;
    except
    end;
  end;
  inherited GotoBookmark(aBookmark);
end;

procedure TanADOPipeline.FreeBookmark(aBookmark: LongInt);
begin
//  inherited FreeBookmark(aBookmark);
end;

function TanADOPipeline.GetDataSetName: String;
begin
  Result := Name;
end;

function TanADOPipeline.GetFieldNames(aFieldNameList: TStrings): Boolean;
var
  i: Integer;
  Res: Boolean;
begin
  Res := False;
  if Assigned(aFieldNameList) then
  begin
    aFieldNameList.Clear;
    try
      if (VarType(FRecordset) = varEmpty) or (VarType(FRecordset) = varNull) then
      begin
        // needed at designtime
        OpenDataset;
      end;
      if (VarType(FRecordset) = varDispatch) and (FRecordset.State <> adStateClosed) then
      begin
        for i := 0 to FRecordset.Fields.Count - 1 do
        begin
          aFieldNameList.Add(FRecordset.Fields.Item[i].Name);
        end;
        Res := True;
      end;
    except
      aFieldNameList.Clear;
      Res := False;
    end;
  end;
  Result := Res;
end;

function TanADOPipeline.GetFieldDataType(
  aFieldName: String): TppDataType;
var
  Res: TppDataType;
begin
  // I am not sure that I have made the right choices for all
  // possible datatypes here but it seems to work ok. If
  // someone find any problems please contact me.
  Res := dtVariant;
  if (VarType(FRecordset) = varEmpty) or (VarType(FRecordset) = varNull) then
  begin
    // needed at designtime
    OpenDataset;
  end;
  if (VarType(FRecordset) = varDispatch) and (FRecordset.State <> adStateClosed) then
  begin
    try
      case FRecordset.Fields.Item[aFieldName].Type of
        adEmpty:
        begin
          Res := dtVariant;
        end;
        adTinyInt, adSmallInt, adInteger, adUnsignedTinyInt, adUnsignedSmallInt, adUnsignedInt:
        begin
          Res := dtInteger;
        end;
        adBigInt, adUnsignedBigInt:
        begin
          Res := dtLongint;
        end;
        adSingle:
        begin
          Res := dtSingle;
        end;
        adDouble, adDecimal, adNumeric:
        begin
          Res := dtDouble;
        end;
        adCurrency:
        begin
          Res := dtCurrency;
        end;
        adBoolean, adError:
        begin
          Res := dtBoolean;
        end;
        adUserDefined, adVariant, adIDispatch, adIUnknown, adGUID:
        begin
          Res := dtString;
        end;
        adDate, adDBDate:
        begin
          Res := dtDate;
        end;
        adDBTime, adDBTimeStamp:
        begin
          Res := dtDateTime;
        end;
        adBSTR, adChar, adVarChar, adLongVarChar, adWChar, adVarWChar, adLongVarWChar:
        begin
          Res := dtString;
        end;
        adBinary, adVarBinary, adLongVarBinary, adChapter:
        begin
          Res := dtBlob;
        end;
        adFileTime, adDBFileTime:
        begin
          Res := dtDateTime;
        end;
        adPropVariant:
        begin
          Res := dtBlob;
        end;
        adVarNumeric:
        begin
          Res := dtDouble;
        end;
      end;
    except
      Res := dtVariant;
    end;
  end;
  Result := Res;
end;

function TanADOPipeline.GetFieldSize(aFieldName: String): Integer;
var
  Res: Integer;
begin
  Res := 0;
  if (VarType(FRecordset) = varDispatch) and (FRecordset.State <> adStateClosed) then
  begin
    Res := FRecordset.Fields.Item[aFieldName].DefinedSize;
  end;
  Result := Res;
end;

function TanADOPipeline.GetFieldAsDouble(aFieldName: String): Double;
var
  Res: Double;
begin
  Res := 0;
  if (VarType(FRecordset) = varDispatch) and (FRecordset.State <> adStateClosed) then
  begin
    try
      Res := Double(FRecordset.Fields.Item[aFieldName].Value);
    except
      Res := 0;
    end;
  end;
  Result := Res;
end;

function TanADOPipeline.GetFieldAsPicture(
  aFieldName: String): TPicture;
var
  Blob: Variant;
  BlobPtr: Pointer;
  Stream: TMemoryStream;
  Res: TPicture;
begin
  Res := nil;
  Stream := nil;
  if (VarType(FRecordset) = varDispatch) and (FRecordset.State <> adStateClosed) then
  begin
    try
      Blob := FRecordset.Fields.Item[aFieldName].Value;
      if VarIsArray(Blob) then
      begin
        BlobPtr := VarArrayLock(Blob);
        Stream := TMemoryStream.Create;
        Stream.Write(BlobPtr^, VarArrayHighBound(Blob, 1) + 1);
        VarArrayUnLock(Blob);
        Blob := NULL;

        Res := TPicture.Create;
        Stream.Position := 0;
        Res.Bitmap.LoadFromStream(Stream);
        if Assigned(Stream) then
        begin
          Stream.Free;
        end;
      end;
    except
      if Assigned(Stream) then
      begin
        Stream.Free;
      end;
      Blob := Null;
      Res := nil;
    end;
  end;
  Result := Res;
end;

function TanADOPipeline.GetFieldAsString(aFieldName: String): String;
begin
  if (VarType(FRecordset) = varDispatch) and (FRecordset.State <> adStateClosed) then
  begin
    try
      Result := String(FRecordset.Fields.Item[aFieldName].Value);
    except
      Result := '';
    end;
  end;
end;

function TanADOPipeline.GetFieldDisplayWidth(
  aFieldName: String): Integer;
var
  Res: Integer;
begin
  Res := 10;
  // The different field sizes may need adjustments
  if (VarType(FRecordset) = varDispatch) and (FRecordset.State <> adStateClosed) then
  begin
    try
      case FRecordset.Fields.Item[aFieldName].Type of
        adEmpty:
        begin
          Res := 0;
        end;
        adTinyInt, adSmallInt, adInteger, adUnsignedTinyInt, adUnsignedSmallInt, adUnsignedInt, adBigInt, adUnsignedBigInt:
        begin
          Res := 15;
        end;
        adSingle, adDouble, adDecimal, adNumeric, adVarNumeric:
        begin
          Res := 15;
        end;
        adCurrency:
        begin
          Res := 15;
        end;
        adBoolean, adError:
        begin
          Res := 5;
        end;
        adUserDefined, adVariant, adIDispatch, adIUnknown:
        begin
          Res := 10;
        end;
        adGUID:
        begin
          Res := 25;
        end;
        adDate, adDBDate, adDBTime, adDBTimeStamp, adFileTime, adDBFileTime:
        begin
          Res := 10;
        end;
        adBSTR, adChar, adVarChar, adLongVarChar, adWChar, adVarWChar, adLongVarWChar:
        begin
          Res := FRecordset.Fields.Item[aFieldName].DefinedSize;
        end;
        adBinary, adVarBinary, adLongVarBinary, adChapter:
        begin
          Res := 10;
        end;
        adPropVariant:
        begin
          Res := 10;
        end;
      end;
    except
      Res := 10;
    end;
  end;
  Result := Res;
end;

function TanADOPipeline.GetFieldIsCalculated(
  aFieldName: String): Boolean;
begin
  // This is not currently used in RB4.06. It should return true
  // if a field is based on calculations in SQL but I dont see any simple
  // ways to implement this. If someone want this to return something else,
  // and have an idea of how to implement it, please contact me.
  Result := False;
end;

function TanADOPipeline.GetFieldValue(aFieldName: String): Variant;
var
  Res: Variant;
begin
  Res := '';
  if (VarType(FRecordset) = varDispatch) and (FRecordset.State <> adStateClosed) then
  begin
    try
      Res := FRecordset.Fields.Item[aFieldName].Value;
    except
      Res := Null;
    end;
  end;
  Result := Res;
end;

procedure TanADOPipeline.RestoreFromCache(aCachePageNo: Integer);
var
  lRec: TppDataPipelineSaveRec;
begin

  ReadRecordFromCache(aCachePageNo, lRec);

  ExcludeState([ppdaNoRecords, ppdaLastRecord, ppdaFirstRecord]);

  if lRec.FNoRecords then IncludeState([ppdaNoRecords]);
  if lRec.FLastRecord then IncludeState([ppdaLastRecord]);
  if lRec.FFirstRecord then IncludeState([ppdaFirstRecord]);

  TraversalCount := lRec.FTraversalCount;

  {restore record position}
  if (ppdaFirstRecord in State) then
    First

  else if (ppdaLastRecord in State) then
    Last

  else if lRec.FBookmarked then
    GotoBookmark(lRec.FBookmark);
end;

procedure TanADOPipeline.SaveToCache(aCachePageNo: Integer);
var
  lRec: TppDataPipelineSaveRec;
begin

  inherited SaveToCache(aCachePageNo);

  {save record states}
  lRec.FNoRecords   := (ppdaNoRecords in State);
  lRec.FLastRecord  := (ppdaLastRecord in State);
  lRec.FFirstRecord := (ppdaFirstRecord in State);

  lRec.FTraversalCount := TraversalCount;

  {bookmark record position}
  if not(ppdaNoRecords in State) and GetActive and
     not(CheckBOF) and not(CheckEOF) then
    begin
      lRec.FBookmarked := True;

      lRec.FBookmark := GetBookmark;
    end
  else
    begin
      lRec.FBookmarked := False;

      lRec.FBookmark := 0;
    end;

  WriteRecordToCache(aCachePageNo, lRec);
end;

procedure TanADOPipeline.Delete;
begin
  if (VarType(FRecordset) = varDispatch) and (FRecordset.State <> adStateClosed) then
  begin
    try
      FRecordset.Delete(adAffectCurrent);
    except
    end;
  end;
end;

procedure TanADOPipeline.Edit;
begin
  if (VarType(FRecordset) = varDispatch) and (FRecordset.State <> adStateClosed) then
  begin
    try
      FRecordset.Update(EmptyParam, EmptyParam);
    except
    end;
  end;
end;

procedure TanADOPipeline.Insert;
begin
  if (VarType(FRecordset) = varDispatch) and (FRecordset.State <> adStateClosed) then
  begin
    try
      FRecordset.AddNew(EmptyParam, EmptyParam);
    except
    end;
  end;
end;

procedure TanADOPipeline.Post;
begin
  if (VarType(FRecordset) = varDispatch) and (FRecordset.State <> adStateClosed) then
  begin
    try
      FRecordset.Update(EmptyParam, EmptyParam);
    except
    end;
  end;
end;

function TanADOPipeline.Locate(const aFieldName: string;
  aKeyValue: Variant; aOptions: TppLocateOptions): Boolean;
var
  Res: Boolean;
  Criteria: String;
begin
  // This function does not make casesensitive searches!!!
  Res := False;
  if (VarType(FRecordset) = varDispatch) and (FRecordset.State <> adStateClosed) then
  begin
    try
      // is searchfield string type ?
      case FRecordset.Fields.Item[aFieldName].Type of
        adBSTR, adChar, adVarChar, adLongVarChar, adWChar, adVarWChar, adLongVarWChar:
        begin
          if pploPartialKey in aOptions then
          begin
            // Partial search
            Criteria := aFieldName + ' LIKE ''*' + aKeyValue + '*''';
          end
          else
          begin
            Criteria := aFieldName + ' LIKE ''' + aKeyValue + '''';
          end;
        end;
      else
        begin
          Criteria := aFieldName + ' = ' + aKeyValue;
        end;
      end;
      FRecordset.MoveFirst;
      FRecordset.Find(Criteria, 0, adSearchForward, EmptyParam);
      if FRecordset.EOF or FRecordset.BOF then
      begin
        Res := False;
      end
      else
      begin
        Res := True;
      end;
    except
      Res := False;
    end;
  end;
  Result := Res;
end;

procedure TanADOPipeline.SetFieldValue(aFieldName: String;
  aValue: Variant);
begin
  if (VarType(FRecordset) = varDispatch) and (FRecordset.State <> adStateClosed) then
  begin
    try
      FRecordset.Fields.Item[aFieldName].Value := aValue;
    except
    end;
  end;
end;

procedure TanADOPipeline.GetFieldAsStream(aFieldName: String;
  aStream: TStream);
var
  Blob: Variant;
  BlobPtr: Pointer;
begin
  if Assigned(aStream) then
  begin
  if (VarType(FRecordset) = varDispatch) and (FRecordset.State <> adStateClosed) then
    begin
      try
        Blob := FRecordset.Fields.Item[aFieldName].Value;
        if VarIsArray(Blob) then
        begin
          BlobPtr := VarArrayLock(Blob);
          aStream.Write(BlobPtr^, VarArrayHighBound(Blob, 1) + 1);
          VarArrayUnLock(Blob);
          Blob := NULL;
        end;
      except
      end;
    end;
  end;
end;

procedure TanADOPipeline.SetFieldFromStream(aFieldName: String;
  aStream: TStream);
var
  Blob: Variant;
  BlobPtr: Pointer;
begin
  if (VarType(FRecordset) = varDispatch) and (FRecordset.State <> adStateClosed) then
  begin
    try
      Blob := VarArrayCreate([1,aStream.Size], varByte);
      BlobPtr := VarArrayLock(Blob);
      aStream.Read(BlobPtr^, aStream.Size);
      VarArrayUnLock(Blob);
      FRecordset.Fields.Item[aFieldName].Value := Blob;
      Blob := NULL;
    except
    end;
  end;
end;

initialization

  RegisterClasses([TanADOPipeline, TanChildADOPipeline]);

finalization

  UnRegisterClasses([TanADOPipeline, TanChildADOPipeline]);

end.
