unit anDBADO;

(*******************************************************************************
v1.1

This component is the result of a project where I needed access to MS SQL
Server with ADO, and is provided as is. It works in my project, but I don't make
any guarantees that it will work for you. You may use this component on your own
risk. I don't take any responsibility for any problems arising from the use
of it (including but not limited to data loss).

Se readme.txt for help with this component and license.txt for conditions of
use.

Anders Nolberger
Sekonden IT AB
Teknikringen 6
583 30 Linkping
Sweden

tel: +46 13 214440
fax: +46 13 214441
Email: Anders.Nolberger@sekonden.se
WEB: HTTP://www.sekonden.se

*******************************************************************************)

interface

uses Classes, SysUtils, Forms, ExtCtrls, DB, DIALOGS,
     ppClass, ppDBPipe, ppDB, ppClasUt,
     daDB, daQuery, daDataVw, anADODatPrv, ppEndUsr,
     ppTypes,
     anADOPipeline, ADODB_TLB; {this is the unit containing the ADOPipeline}



type

  {ADO DataView Classes:

     1.  TanADOSession
           - descendant of TppSession
           - implements GetDatabaseNames, GetTableNames, etc.

     2.  TanADODataSet
          - descendant of TppDataSet
          - implements GetFieldNames for SQL

     3.  TanADOQueryDataView
          - descendant of TppQueryDataView
          - uses the above classes to create the required
            Pipeline -> Report connection
          - uses the TdaSQL object built by the QueryWizard to assign
            SQL to the Conecction and Recordset.

      }

  { TanADOSession }
  TanADOSession = class(TdaSession)
    private

    public

      class function ClassDescription: String; override;
      class function DataSetClass: TdaDataSetClass; override;

//      function  GetAliasDriverName(const aAlias: String): String; override;
      procedure GetDatabaseNames(aList: TStrings); override;
      procedure GetTableNames(const aDatabaseName: String; aList: TStrings); override;
      function  GetDatabaseType(const aDatabaseName: String): TppDatabaseType; override;
      function  ValidDatabaseTypes: TppDatabaseTypes; override;

  end; {class, TanADOSession}



  { TanADODataSet }
  TanADODataSet = class(TdaDataSet)
    private
      FDBCon: Connection;
      Frs: Recordset;
      FConnectionString: String;
      FTableName: String;

      function ConvertADODataType(ADOType: Variant): TppDataType;
    protected
      procedure BuildFieldList; override;
      function  GetActive: Boolean; override;
      procedure SetActive(Value: Boolean); override;
      procedure SetDatabaseName(const aDatabaseName: String); override;
      procedure SetDataName(const aDataName: String); override;

    public
      constructor Create(aOwner: TComponent); override;
      destructor Destroy; override;

      class function ClassDescription: String; override;

      procedure GetFieldNamesForSQL(aList: TStrings; aSQL: TStrings); override;
      procedure GetFieldsForSQL(aList: TList; aSQL: TStrings); override;

  end; {class, TanADODataSet}


  { TanADOQueryDataView }
  TanADOQueryDataView = class(TdaQueryDataView)
    protected
      procedure Loaded; override;
      procedure UpdateQueryObject;

      procedure SetReport(aReport: TppCustomReport); override;
      procedure SetActive(Value: Boolean); override;

      procedure SQLChanged; override;

      function GetSQLQuery: String;
      function GetConnectionString: String;
    public
      constructor Create(aOwner: TComponent); override;
      destructor Destroy; override;

      class function PreviewFormClass: TFormClass; override;
      class function SessionClass: TClass; override;

      procedure Init; override;

    published
     property SQLQuery: String read GetSQLQuery;
     property ConnectionString: String read GetConnectionString;

  end; {class, TanADOQueryDataView}

  {Delphi design time registration}
  procedure Register;
  function GetRealConnectionstring(Databasename: String): String;
implementation

procedure Register;
begin
  {ADO DataViews}
  RegisterNoIcon([TanADOQueryDataView]);
end;

{ TanADOSession }

class function TanADOSession.ClassDescription: String;
begin
  Result := 'ADOSession';
end;

class function TanADOSession.DataSetClass: TdaDataSetClass;
begin
  Result := TanADODataSet;
end;

{
function TanADOSession.GetAliasDriverName(const aAlias: String): String;
begin
  Result := '';
end;
}

procedure TanADOSession.GetDatabaseNames(aList: TStrings);
var
  Names: TStringList;
  Form: TForm;
  i, j: Integer;
begin
  Names := TStringList.Create;

  for i := 0 to Application.ComponentCount - 1 do
  begin
    if Application.Components[i] is TForm then
    begin
      Form := TForm(Application.Components[i]);
      for j := 0 to Form.ComponentCount - 1 do
      begin
        if Form.Components[j].ClassName = 'TanADOPipeline' then
        begin
          if TanADOPipeline(Form.Components[j]).Template then
          begin
            Names.Add(TanADOPipeline(Form.Components[j]).UserName);
          end;
        end;
      end;
    end;
  end;

  aList.Assign(Names);
  Names.Free;
end;

procedure TanADOSession.GetTableNames(const aDatabaseName: String;
  aList: TStrings);
var
  Tables: TStringList;
  dbCon: Connection;
  rs: Recordset;
begin
  Tables := nil;
  try
    Tables := TStringList.Create;
    dbCon := CoConnection.Create;
    if Assigned(dbCon) then
    begin
      dbCon.CursorLocation := adUseClient;
      dbCon.ConnectionTimeout := 10;
      dbCon.Open(GetRealConnectionstring(aDatabaseName), '', '', 0);
      if dbCon.State = adStateOpen then
      begin
        rs := dbCon.OpenSchema(adSchemaTables, EmptyParam, EmptyParam);

        if Assigned(rs) then
        begin
          while not rs.EOF do
          begin
            try
              // msysmodules and msysmodules2 is system tables from Accesss that
              // is not marked as SYSTEM TABLE but I dont want them to be visible
              // If you want to use these names in another database just delete
              // the two rows below
              if (rs.Fields.Item['TABLE_TYPE'].Value <> 'SYSTEM TABLE')
                and (rs.Fields.Item['TABLE_TYPE'].Value <> 'SYSTEM VIEW')
                and (LowerCase(rs.Fields.Item['TABLE_NAME'].Value) <> 'msysmodules')
                and (LowerCase(rs.Fields.Item['TABLE_NAME'].Value) <> 'msysmodules2') then
              begin
                Tables.Add(rs.Fields.Item['TABLE_NAME'].Value);
              end;
            except
            end;
            rs.MoveNext;
          end;
        end;
      end
    end;
  finally
    if Assigned(dbCon) then
    begin
      if (dbCon.State and adStateOpen) = adStateOpen then
      begin
        dbCon.Close;
      end;
      dbCon := nil;
    end;
    aList.Assign(Tables);
    Tables.Free;
  end;
end;

function  TanADOSession.GetDatabaseType(const aDatabaseName: String): TppDatabaseType;
begin
  Result := dtMSSQLServer;
end;

function  TanADOSession.ValidDatabaseTypes: TppDatabaseTypes;
begin
  Result := [dtAdvantage, dtInterBase, dtMSAccess, dtMSSQLServer,
             dtSybaseASE, dtSybaseASA, dtOracle, dtParadox, dtDBISAM, dtOther];
end;

{ TanADODataSet }

procedure TanADODataSet.BuildFieldList;
var
  lField: TppField;
  tmpName: String;
  i: Integer;
begin
  inherited BuildFieldList;

  try
    if not Self.getActive then
    begin
      Self.SetActive(True);
    end;
  except
    Exit;
  end;

  if Assigned(Frs) then
  begin
    if (Frs.State and adStateOpen) = adStateOpen then
    begin
{      if not(Frs.EOF and Frs.BOF) then
      begin
        Frs.MoveFirst;
        while not Frs.EOF do
        begin
          lField := TppField.Create(nil);
          lField.TableName := FTableName;
          lField.FieldName := Frs.Fields.Item['COLUMN_NAME'].Value;
          lField.DataType := ConvertADODataType(Frs.Fields.Item['DATA_TYPE'].Value);
          AddField(lField);

          Frs.moveNext;
        end;
      end;   }
      for i := 0 to Frs.Fields.Count - 1 do
      begin
        lField := TppField.Create(nil);
        lField.TableName := FTableName;
        tmpName := Frs.Fields.Item[i].Name;
        if Pos('.', tmpName) = 0 then
        begin
          lField.Fieldname := tmpName;
        end
        else
        begin
          lField.Fieldname := Copy(tmpName, Pos('.', tmpName) + 1, Length(tmpName));
        end;
        lField.DataType := ConvertADODataType(Frs.Fields.Item[i].Type_);
        AddField(lField);
      end;
    end;
  end;
end;

class function TanADODataSet.ClassDescription: String;
begin
  Result := 'ADODataset';
end;

function TanADODataSet.ConvertADODataType(ADOType: Variant): TppDataType;
var
  Res: TppDataType;
begin
  Case ADOType of
    adEmpty:
    begin
      Res := dtVariant;
    end;
    adTinyInt, adSmallInt, adInteger, adUnsignedTinyInt, adUnsignedSmallInt, adUnsignedInt:
    begin
      Res := dtInteger;
    end;
    adBigInt, adUnsignedBigInt:
    begin
      Res := dtLongint;
    end;
    adSingle:
    begin
      Res := dtSingle;
    end;
    adDouble, adDecimal, adNumeric:
    begin
      Res := dtDouble;
    end;
    adCurrency:
    begin
      Res := dtCurrency;
    end;
    adBoolean, adError:
    begin
      Res := dtBoolean;
    end;
    adUserDefined, adVariant, adIDispatch, adIUnknown, adGUID:
    begin
      Res := dtString;
    end;
    adDate, adDBDate:
    begin
      Res := dtDate;
    end;
    adDBTime, adDBTimeStamp:
    begin
      Res := dtDateTime;
    end;
    adBSTR, adChar, adVarChar, adLongVarChar, adWChar, adVarWChar, adLongVarWChar:
    begin
      Res := dtString;
    end;
    adBinary, adVarBinary, adLongVarBinary, adChapter:
    begin
      Res := dtBlob;
    end;
    adFileTime, adDBFileTime:
    begin
      Res := dtDateTime;
    end;
    adPropVariant:
    begin
      Res := dtBlob;
    end;
    adVarNumeric:
    begin
      Res := dtDouble;
    end;
  else
    begin
      Res := dtVariant;
    end;
  end;
  Result := Res;
end;

constructor TanADODataSet.Create(aOwner: TComponent);
begin
  inherited Create(aOwner);

  FDBCon := CoConnection.Create;
  Frs := nil;
end;

destructor TanADODataSet.Destroy;
begin
  try
    if Assigned(Frs) then
    begin
      if (Frs.State and adStateOpen) = adStateOpen then
      begin
        Frs.Close;
      end;
      Frs := nil;
    end;
    if Assigned(FDBCon) then
    begin
      if (FDBCon.State and adStateOpen) = adStateOpen then
      begin
        FDBCon.Close;
      end;
      FDBCon := nil;
    end;
  except
  end;

  inherited Destroy;
end;

function TanADODataSet.GetActive: Boolean;
var
  Res: Boolean;
begin
  Res := False;
  try
    if Assigned(Frs) then
    begin
      if (Frs.State and adStateOpen) = adStateOpen then
      begin
        Res := True;
      end;
    end;
  except
    Res := False;
  end;

  Result := Res;
end;

procedure TanADODataSet.GetFieldNamesForSQL(aList, aSQL: TStrings);
var
  i: Integer;
  tmpSQL: String;
  rs: Recordset;
begin
  aList.Clear;

  tmpSQL := '';
  for i := 0 to aSQL.Count - 1 do
  begin
    tmpSQL := tmpSQL + ' ' + aSQL.Strings[i];
  end;

  try
    if Assigned(FDBCon) then
    begin
      if (FDBCon.State and adStateClosed) = adStateClosed then
      begin
        FDBCon.Open(GetRealConnectionstring(FConnectionString), '', '', 0);
      end;
      rs := CoRecordset.Create;
      rs.Open(tmpSQL, FdbCon, adOpenForwardOnly, adLockReadOnly, adCmdUnknown);
      for i := 0 to rs.Fields.Count - 1 do
      begin
        aList.Add(rs.Fields.Item[i].Name);
      end;
      rs.Close;
    end;
  except
  end;
  rs := nil;
end;

procedure TanADODataSet.GetFieldsForSQL(aList: TList; aSQL: TStrings);
var
  i: Integer;
  tmpSQL: String;
  rs: Recordset;
  lField: TppField;
begin
  aList.Clear;

  tmpSQL := '';
  for i := 0 to aSQL.Count - 1 do
  begin
    tmpSQL := tmpSQL + ' ' + aSQL.Strings[i];
  end;

  try
    if Assigned(FDBCon) then
    begin
      if (FDBCon.State and adStateClosed) = adStateClosed then
      begin
        FDBCon.Open(GetRealConnectionstring(FConnectionString), '', '', 0);
      end;
      rs := CoRecordset.Create;
      rs.Open(tmpSQL, FdbCon, adOpenForwardOnly, adLockReadOnly, adCmdUnknown);
      for i := 0 to rs.Fields.Count - 1 do
      begin
        lField := TppField.Create(nil);
        lField.Fieldname := rs.Fields.Item[i].Name;
        lField.DataType := ConvertADODataType(rs.Fields.Item[i].Type_);
        aList.Add(lField);
      end;
      rs.Close;
    end;
  except
  end;
  rs := nil;
end;

procedure TanADODataSet.SetActive(Value: Boolean);
var
  Opt: Variant;
begin
  try
    if Value then
    begin
      if Assigned(FDBCon) then
      begin
        if (FDBCon.State and adStateOpen) <> adStateOpen then
        begin
          FDBCon.Open(GetRealConnectionstring(FConnectionString), '', '', 0);
        end;
        Opt := VarArrayCreate([1,4], varVariant);
        opt[3] := FTableName;
        Frs := CoRecordset.Create;
        if Assigned(Frs) then
        begin
          Frs.open('SELECT TOP 1 * FROM ' + FTableName, FDBCon, adOpenStatic, adLockReadOnly, adCmdText);
        end;
      end;
    end
    else
    begin
      if Assigned(Frs) then
      begin
        if (Frs.State and adStateOpen) = adStateOpen then
        begin
          Frs.Close;
        end;
        Frs := nil;
      end;
    end;
  except
  end;
end;

procedure TanADODataSet.SetDatabaseName(const aDatabaseName: String);
begin
  inherited SetDatabaseName(aDatabaseName);
  Self.SetActive(False);
  FConnectionString := aDatabaseName;
end;

procedure TanADODataSet.SetDataName(const aDataName: String);
begin
  inherited SetDataName(aDataName);
  Self.SetActive(False);
  FTableName := aDataName;
end;

{ TanADOQueryDataView }

constructor TanADOQueryDataView.Create(aOwner: TComponent);
begin
  inherited Create(aOwner);
end;

destructor TanADOQueryDataView.Destroy;
begin
  inherited Destroy;
end;

procedure TanADOQueryDataView.Init;
var
  lDataPipeline: TanChildADOPipeline;
begin
  inherited Init;

  if DatapipelineCount > 0 then Exit;

  lDataPipeline := TanChildADOPipeline(ppComponentCreate(Self, TanChildADOPipeline));
  lDataPipeline.AutoCreateFields := False;
  lDataPipeline.DataView := Self;
end;

procedure TanADOQueryDataView.Loaded;
begin
  inherited Loaded;

  UpdateQueryObject;
  SetActive(true);
end;

class function TanADOQueryDataView.PreviewFormClass: TFormClass;
begin
  Result := TFormClass(GetClass('TanADOPreviewDataDialog'));
//  Result := TFormClass(GetClass(''));
end;

class function TanADOQueryDataView.SessionClass: TClass;
begin
  Result := TanADOSession;
end;

procedure TanADOQueryDataView.SetActive(Value: Boolean);
begin
  inherited SetActive(Value);

  if (DataPipelineCount = 0) then Exit;

  if DataPipelines[0].Active <> Value then
    DataPipelines[0].Open;
end;

procedure TanADOQueryDataView.SetReport(aReport: TppCustomReport);
begin
  inherited SetReport(aReport);

  if (Report = nil) or (DataPipelineCount = 0) then Exit;

  Report.DataPipeline := DataPipelines[0];
end;

procedure TanADOQueryDataView.SQLChanged;
begin
  UpdateQueryObject;
end;

procedure TanADOQueryDataView.UpdateQueryObject;
begin

  if (DataPipelineCount = 0) then Exit;

  TanADOPipeline(DataPipelines[0]).ConnectionString := SQL.Databasename;
  TanADOPipeline(DataPipelines[0]).ConStringIsShortcut := True;
  TanADOPipeline(DataPipelines[0]).SQLQuery := GetSQLQuery;
end;

function TanADOQueryDataView.GetConnectionString: String;
begin
  Result := SQL.Databasename;
end;

function TanADOQueryDataView.GetSQLQuery: String;
var
  i: Integer;
  tmpSQL: String;
begin
  tmpSQL := '';
  for i := 0 to SQL.SQLText.Count - 1 do
  begin
    tmpSQL := tmpSQL + ' ' + SQL.SQLText.Strings[i];
  end;
  Result := tmpSQL;
end;

function GetRealConnectionstring(Databasename: String): String;
var
  i, j: Integer;
  Form: TForm;
  ConStr: String;
begin
  ConStr := Databasename;
  for i := 0 to Application.ComponentCount - 1 do
  begin
    if Application.Components[i] is TForm then
    begin
      Form := TForm(Application.Components[i]);
      for j := 0 to Form.ComponentCount - 1 do
      begin
        if Form.Components[j].ClassName = 'TanADOPipeline' then
        begin
          if TanADOPipeline(Form.Components[j]).UserName = Databasename then
          begin
            ConStr := TanADOPipeline(Form.Components[j]).Connectionstring;
          end;
        end;
      end;
    end;
  end;
  Result := ConStr;
end;

initialization
  {register DADE descendant session, dataset, dataview}
  daRegisterSession(TanADOSession);
  daRegisterDataSet(TanADODataSet);
  daRegisterDataView(TanADOQueryDataView);

finalization
  {unregister DADE descendant the session, dataset, dataview}
  daUnRegisterSession(TanADOSession);
  daUnRegisterDataSet(TanADODataSet);
  daUnRegisterDataView(TanADOQueryDataView);

end.
